{
  "type": "record",
  "Aspect": {
    "name": "formSettings"
  },
  "name": "FormSettings",
  "namespace": "com.linkedin.pegasus2avro.form",
  "fields": [
    {
      "type": {
        "type": "record",
        "name": "FormNotificationSettings",
        "namespace": "com.linkedin.pegasus2avro.form",
        "fields": [
          {
            "Searchable": {
              "fieldType": "BOOLEAN"
            },
            "type": "boolean",
            "name": "notifyAssigneesOnPublish",
            "default": false,
            "doc": "Whether to notify the assignees of the form when it is published."
          }
        ]
      },
      "name": "notificationSettings",
      "doc": "Notification settings for a form"
    }
  ]
}