# mypy: ignore-errors
# flake8: noqa

# This file is autogenerated by /metadata-ingestion/scripts/avro_codegen.py
# Do not modify manually!

# pylint: skip-file
# fmt: off
# isort: skip_file
from .....schema_classes import CorpGroupEditableInfoClass
from .....schema_classes import CorpGroupInfoClass
from .....schema_classes import CorpGroupSettingsClass
from .....schema_classes import CorpUserAppearanceSettingsClass
from .....schema_classes import CorpUserCredentialsClass
from .....schema_classes import CorpUserEditableInfoClass
from .....schema_classes import CorpUserHomePageSettingsClass
from .....schema_classes import CorpUserInfoClass
from .....schema_classes import CorpUserInvitationStatusClass
from .....schema_classes import CorpUserSettingsClass
from .....schema_classes import CorpUserStatusClass
from .....schema_classes import CorpUserViewsSettingsClass
from .....schema_classes import GroupMembershipClass
from .....schema_classes import InvitationStatusClass
from .....schema_classes import InviteTokenClass
from .....schema_classes import NativeGroupMembershipClass
from .....schema_classes import RoleMembershipClass
from .....schema_classes import TokenTypeClass
from .....schema_classes import UserAiPluginConfigClass
from .....schema_classes import UserAiPluginSettingsClass
from .....schema_classes import UserApiKeyConnectionConfigClass
from .....schema_classes import UserOAuthConnectionConfigClass


CorpGroupEditableInfo = CorpGroupEditableInfoClass
CorpGroupInfo = CorpGroupInfoClass
CorpGroupSettings = CorpGroupSettingsClass
CorpUserAppearanceSettings = CorpUserAppearanceSettingsClass
CorpUserCredentials = CorpUserCredentialsClass
CorpUserEditableInfo = CorpUserEditableInfoClass
CorpUserHomePageSettings = CorpUserHomePageSettingsClass
CorpUserInfo = CorpUserInfoClass
CorpUserInvitationStatus = CorpUserInvitationStatusClass
CorpUserSettings = CorpUserSettingsClass
CorpUserStatus = CorpUserStatusClass
CorpUserViewsSettings = CorpUserViewsSettingsClass
GroupMembership = GroupMembershipClass
InvitationStatus = InvitationStatusClass
InviteToken = InviteTokenClass
NativeGroupMembership = NativeGroupMembershipClass
RoleMembership = RoleMembershipClass
TokenType = TokenTypeClass
UserAiPluginConfig = UserAiPluginConfigClass
UserAiPluginSettings = UserAiPluginSettingsClass
UserApiKeyConnectionConfig = UserApiKeyConnectionConfigClass
UserOAuthConnectionConfig = UserOAuthConnectionConfigClass

# fmt: on
